####################
# version 2.0
####################




"""
                                                                              
NAME Bio.seqDB
DESC Sequences Database

Facilities to export sequences from chromosome files

classes:
    
methods:
    is_RefSeq
    read_slice
    process_row
    get_sequence
    get_position
    get_promoter
                                                                              
"""



# IMPORTS
#-------------------------------------------------------------------------------
import sys
import os
import random


import Core.UI.cli as cli 
import Core.Types.list as list
import Core.Persistence.op as op

import sequence


#-------------------------------------------------------------------------------
# IMPORTS



# CODE
#-------------------------------------------------------------------------------

CHR_EXTENSION = '.raw'

def is_RefSeq(id):
    """
    return true if id is a RefSeq identifier, wrong if not

    id -- a string as identifier
    """
    if id.startswith('NM_'):
        return True
    else:
        return False




def read_slice(f, start, end):
    """
    Read a slice of bytes in file
    NOTE: start and end included, index start at 0
    f must open in binary mode

    f     -- file
    
    start -- start of the slice
    
    end   -- end of the slice
    """
    assert('b' in f.mode)
    try:
        f.seek(start)
        return f.read(end-start+1)
    except IOError:
	 return ''




def process_row(row):
    """
    return the information extracted from a row string

    row    -- a string with the info about a promoter sequence
    
    return -- the chromosom where the sequence is, the strand and the TSS
    """
    chr, strand, exonStarts, exonEnds = row

    exonStarts = exonStarts.split(',')[:-1]
    exonEnds = exonEnds.split(',')[:-1]

    if strand == '+':
        return chr, strand, int(exonStarts[0])
    else:
        #NOTE the -1 : UCSC use [a,b[ intervals
        return chr, strand, int(exonEnds[-1])-1


def get_sequence(chr, i, j, organism, dir):
    """
    Get sequence corresponding to absolute location on assembly
    
    chr      -- string
    
    i, j     -- absolute location [i..j] on assembly j can be < i in this case return reverse complement [j..i]
    NOTE [0...N]
    
    organism -- the organism from which the promoter is taken
    
    dir      -- the directory where to find the sequences of the chromosomes
    
    return   -- a dna sequence
    """
    try:

        f = open(os.path.join(dir, organism, chr+ CHR_EXTENSION), 'rb')
	
        try:
            if j > i:
                dna = read_slice(f, i, j)
                if dna != "":
                    dna=dna.upper()
            else:
                dna=read_slice(f, j, i)
                if dna != "":
                	dna = sequence.reverse_complement(dna.upper())
        finally:
            f.close()
        return dna
    except IOError:
        print >> sys.stderr, 'IO Error while reading chr data'
        return ''


def get_position(organism, chr, strand, TSS, location):
    """
    return the relative position of the promoter from the TSS

    organism -- the organism from which the promoter is taken
    
    chr      -- the chromosome where the promoter sequence is
    
    strand   -- the strand of chromosome on which the promoter sequence is
    
    TSS      -- the position of the transcirption start site
    
    location -- the absolute location of the promoter on the chromosome
    """
    start, end = location
    if strand == '+':
        a,b = TSS + start, TSS + end
        i,j = a,b
    else:
        a,b = TSS - start,  TSS - end
        i,j = b,a    
    return chr, a, b, organism


def get_promoter(organism, chr, strand, TSS, location=(-1000,-1), dir=None):
    """
    Return promoter DNA sequence

    organism -- the organism from which the promoter is taken
    
    chr      -- the chromosome where the promoter sequence is
    
    strand   -- the strand of chromosome on which the promoter sequence is
    
    TSS      -- the position of the transcirption start site
    
    location -- the absolute location of the promoter on the chromosome
    
    dir      -- the directory where to find the sequences of the chromosomes
    
    return   -- a dna sequence
    """

    chr, a, b, organism = get_position(organism, chr, strand, TSS, location)
  
    r = get_sequence(chr, a, b, organism, dir)

    if not sequence.is_valid(None, (location[0],location[1]), r):
        return None
    return r


#-------------------------------------------------------------------------------
# CODE





