####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Bio.Types.sequence
# DESC Sequence Objects
#                                                                              #
################################################################################
"""

Objects to handle sequences

classes:
    Sequences
    FastaIterator
    Sequence

methods:
    str2location
    location2str
    is_valid_location
    sequences2dna
    dna2sequences
    reshape_sequences
    sequences2fasta
    fasta2sequences
    find_organism
    sl2sequences
    rnd2sequences
    find_location
    decode_title
    encode_title
    reshaped
    reverse_complement
    is_valid
    gc_percent
    ACGT_freq_dna
    ACGT_freq
    decode_ref
    IUPAC2list
    _extendn
    extendn
    _convert_N
    convert_N
    IUPAC2fulllist
    fasta2dump
"""

# IMPORTS
#-------------------------------------------------------------------------------
import re
import random
import sys

from Core.Types.pipe import Output, Input
import Core.UI.cli as cli
import Core.Persistence.op as op
import Core.types as types
import Core.utils
import fasta
import os
from StringIO import StringIO
import math
import input_checker as ic
from StringIO import StringIO

DATADIR="/home/laurie/Bureau/data-tfme/"
SEQFILE="%s.seq.db"

#-------------------------------------------------------------------------------
# IMPORTS
 
# OBJECTS
#-------------------------------------------------------------------------------
class Sequences(list):
    """
    list of sequences object
    """
    def __init__(self, l=[], name=''):
        super(Sequences, self).__init__(l)
        self.name = name
        self.info  = {}
        self.idx   = None
        #self.location = find_location(self)
        
    def get_location(self):
        """
        return the location of a sequence
        """
        self.location = find_location(self)
        return self.location
        
    def __repr__(self):
        self.get_location()
        return '%s [%d sequences] [%+05d:%+05d]' % (self.name, len(self), self.location[0], self.location[1])

    def create_idx(self):
        """
        create an id to the sequence
        """
        self.idx =  types.list.index(self, lambda s: s.id)
        return self.idx


    def change_locations(self, new_origins):
        """update each sequence location
        """
        assert(len(self)==len(new_origins))
        for i in range(len(self)):
            self[i].location  = (self[i].location[0] - new_origins[i], self[i].location[1] - new_origins[i])


class FastaIterator:
    """Sequence Iterator on Fasta File

    Return Sequence Object
    """

    def __init__(self, f):
        self.it = fasta.FastaReader(f)
        
    
    def next(self):
        """
        extract a sequence from the fasta
        """
        title, dna= self.it.next()
        name, location = decode_title(title)
        return Sequence(name, location, sequence)


class Sequence:
    """Simple sequence object
    """
    def __init__(self, id=None, location=None, sequence='', organism='unknown', autoreshape=False):
        """
        id        -- string ex NM_000043
        
        location  -- (start, end)  start, end: int position (relative to TSS) default=(0, len(sequence)-1)
                     start and end are included
                     
        sequence  -- DNA string 'ATGC...' default=''
        
        NOTE: sequence can be automaticaly reshaped
        """

        self.id       = id
        self.location = location or (0,len(sequence)-1) #relative location
        self.locationTSS = self.location # location relative to TSS
        self.sequence = sequence.upper()
        self.organism = organism

        if autoreshape:
            newLocation = (self.location[0], self.location[0]+len(sequence) - 1)
            if sequence and tuple(newLocation) != tuple(self.location):
                self.reshape(newLocation)
                cli.warning("Sequence '%s' length don't match given values - Auto reshape to [%+.5d:%+.5d]" \
                % (`self`, self.location[0], self.location[1]))

    def reshape(self, newLocation):
        """Reshape sequence to match newLocation (start, end)
        """
        self.locationTSS = self.location
        if not newLocation or self.location == newLocation:
            return self
        self.sequence = reshaped(self.sequence, self.location , newLocation)
        self.location = newLocation

    def reverse_complement(self):
        self.sequence = reverse_complement(self.sequence)

    def gc(self):
        #return 0.5
        return gc_percent(self.sequence)
    
    def tuple(self):
        """
        Return tuple (id, location, sequence)
        """
        return (self.id, self.location, self.sequence)

    def __len__(self):
        return self.location[1] - self.location[0] + 1

    def __repr__(self):
        return encode_title(self.id, self.location)

    def get_dna(self, location=None):
        location = location or self.location
        if location[0] > self.location[1] or location[1] < self.location[0]:
            return ''
        if self.location[0] >= location[0]:
            a = 0
        else:
            a = location[0]-self.location[0]
        if self.location[1] <= location[1]:
            return self.sequence[a:]
        else:
            b = self.location[1] - location[1]
            return self.sequence[a:-b]


#-------------------------------------------------------------------------------
# OBJECTS

# CODE
#-------------------------------------------------------------------------------

def str2location(s):
    """
    convert a location from string to array of int
    """
    try:
        return ( int(s.split(':')[0]), int(s.split(':')[1]) ) 
    except:
        print "Error in the location format"

def location2str(location):
    """
    convert a location from int to string
    """
    return '[%05d:%05d]' % location 

def is_valid_location(start, end, min, max):
    """
    return true if the location is between min and max, false if not

    start -- the start point of the location
    
    end -- the end point of the location
    
    min -- the min boundary for a location
    
    max -- the max boundary for a location
    """
    if start >= end:
        return False
    if start < min or start > max:
        return False
    if end < min or end > max:
        return False
    return True


def sequences2dna(sequences):
    """
    return a list of dna from the sequences
    """
    return [s.sequence for s in sequences]

def dna2sequences(dna, rightPosition=-1):
    """
    dnas -- list of dna strings
    
    return Sequences
    """
    sequences = Sequences()
    i = 1
    for d in dna:
        location = (rightPosition - len(d) + 1, rightPosition)
        sequences.append(Sequence(id=str(i), sequence=d, location=location))
        i +=1
    return sequences


def reshape_sequences(sequences, newLocation):
    """
    Change location of sequences
    """
    for s in sequences:
        s.reshape(newLocation)

def sequences2fasta(sequences, f):
    """
    create a fasta file from the sequences

    sequences -- list of sequences to print in the file
    
    f -- name of the file to create
    """
    try:
        info = cli.Info(len(sequences), 1)
        f = fasta.fasta(f)
        for s in sequences:
            f.append(encode_title(s.id, s.location, s.organism), s.sequence)
	    info('Writing fasta for %s ' % (s.id))
    except IOError:
            cli.error("impossible to create fasta file")
            sys.exit(1)

def fasta2sequences(seq, location, seqDB=None, organism='hg18'):
    """
    Read a fasta file and create a list of Sequence object. 
    change the location if needed
    
    seq      -- a string with a multifasta
    
    location -- the location of the sequences in the fasta
    
    seqDB    -- a database of sequences (used to find the organism id needed)
    
    organism -- the organism by default
    
    return   -- a list of Sequence objects
    """
    try:
        l = []
        assemblies=['rn3', 'mm8','hg18']
        warning=""
        
        
        #uf=StringIO(seq)
        uf=file(seq,'r')
                        
        ##find the smallest dna sequence
        strike=0
        smallestDna=0
        for title1,dna1 in fasta.FastaReader(uf):
            
            if strike == 0:
                smallestDna=len(dna1)
            else:
                if len(dna1) < smallestDna:
                    smallestDna=len(dna1)
            
            strike+=1
         
        ##choose the location
        lenLoc=(location[1]-location[0])+1
        if lenLoc<=smallestDna:
            loc=location
        else:
            loc=[-smallestDna+1,0]
            warning="Some sequences were shorter than the entered location. The location has been changed to "+str(loc)+"\n"
         
         
        uf2=file(seq,'r')
        ##extract the sequences    
        for title, dna in fasta.FastaReader(uf2):
            
            dna = dna.upper()
            id,org,other=decode_title(title)
             
            lenLocation=(loc[1]-loc[0])+1
            removePart=len(dna)-lenLocation
            if removePart >0:
                dna=dna[removePart:]

            if is_valid(id, loc, dna):
                
                l.append( Sequence(id, loc, dna, organism=org or find_organism(id,seqDB)) )

            else:
                cli.warning('Invalid sequence %s' % (title))
        r = Sequences(l)
        r.get_location()
        return (r,warning)
    except:
        cli.error('Error while reading input sequence')
        raise
        sys.exit(1)


def find_organism(key, databases, removeEmpty=1, default=None):
    """
    find the organism of a sequence from its id.
    
    key       -- sequence id
    
    databases -- list of possible organisms
    
    return -- the organism found, otherwise 'hg18'
    """
    if databases is not None:
        assert(type(databases) == list or type(databases) == tuple)
    
        for j in databases:
       
                x = j.get(key, default)          
                if not(not x and removeEmpty):
                    return x.organism
    else:
        return 'hg18'

    
    
def sl2sequences(uf, location, sDB):
    """
    Read a list of sequence identifiers and retreive the info in a database
    remove bad sequences if no id found ...)
    
    uf -- file with the list of identifiers
    
    location -- the boundaries of dna sequence to extract
    
    sDB -- a database of sequences
    
    return -- a list of Sequence object
    """
    sl = types.sl.uf2sl(uf)
    warning=""
    
    sequences  = op.get_list(sl, sDB, removeEmpty=1)
    
    seqFin=[s.id for s in sequences]
    diff=filter(lambda x:x not in seqFin,sl)
    if len(diff) > 0:
        warning="The following RefSeq identifiers were not available: "
        for d in diff:
            warning+=d+","
    
    [ s.reshape( location ) for s in sequences ]
    
    return (Sequences(sequences, sl.title),warning)

def l2sequences(l, location, sDB):
    """
    remove bad sequences (no id found ...)
    out of date
    """
    sequences = op.get_list(sl, sDB, removeEmpty=1)
    [ s.reshape( location ) for s in sequences ]
    return Sequences(sequences, '?')

def markov2sequences(mm, location, organism, n):
    """
    Generate sequences from a Markov model
    
    unused
    """
    return [ Sequence(id='%d' % i, location=location, sequence=mm.generate(location[1]-location[0]+1), organism=organism) for i in range(n) ]

def rnd2sequences(n, location, sDB):
    """
    Return n random sequences from a database
    
    n -- number of sequences to generate
    
    location -- the boundaries of dna sequence to extract
    
    sDB -- a sequences database
    
    return -- a Sequences object
    """
    l = sDB.keys()
    random.shuffle(l)
    l = l[:int(n+n*0.1)]
    sequences = []
    sequences = [s for s in sDB.get_list(l) if s][:n]
    [ s.reshape( location ) for s in sequences ]
    return Sequences(sequences, 'rnd %d ' % (n) )


def find_location(sequences):
    """
    compare the location of several sequences and find the smallest
    
    sequences -- a list of Sequence object
    
    return -- a tuple with 2 floats
    """
    if len(sequences) > 0:
        return (max([i.location[0] for i in sequences]), min([i.location[1] for i in sequences]))
    else:
        return 0,0
#
# Title 
#
def decode_title(title):
    """
    Decode headers of a fasta file
    
    title  -- a title Ex 'NM_000043|human

    return -- tuple (seqid, organism, other)
    """
    try:
        #title = title.split()[0]
 
        split = title.split('|')
       
                
        if len(split) == 2:
            seqid=split[0][1:]
            organism = split[1].strip()
            organism=organism.lower()
            other= None
            
        elif len(split) == 1:
            seqid=split[0][1:]
            organism=None
            other=None      
        elif len(split) >= 3:
            seqid=split[0][1:]
            organism=split[1].strip()
            other= split[2]
            organism=organism.lower()
            #if location.find(':') != -1:
             #   start, end = location.split(':')
             #   location=(int(start), int(end))
           # else:
            #  location="unknown"
        elif len(split) == 0:
            seqid=None
            other=None
            organism=None

        return (seqid, organism, other)
    except:
        return (None,None,None)
        cli.error("the format of a title in the fasta file is not correct. Please check your file ")
	
    



def encode_title(seqid, location=None, organism=None):
    """
    Encode headers for a fasta file
    
    seqid    -- string gene seqid Ex NM_000043
    
    location -- tuple (start, end) 

    return   -- string title
    """
    if not location:
        return '%s' % seqid
    elif not organism:
        return '%s|%+d:%+d' % ( seqid, location[0], location[1])
    else:
        return '%s|%s|%+d:%+d' % ( seqid, str(organism), location[0], location[1])

#
# Shape
#
def reshaped(dna, location, new_location):
    """
    Safe reshape of a dna sequence
    Example reshape_sequence('ATG', [2,4], [2,3]) return 'AT'

    dna          -- string dna
    
    location     -- tuple [start, end]
    
    new_location -- tuple [new_start, start]

    Return       -- reshaped string from [new_start to new_end]
    """
    if new_location[0] > new_location[1]:
        return ''
    return dna[max(new_location[0]-location[0], 0):min(location[1] - location[0]+1, new_location[1] - location[0] +1)]

#
# reverse complement
#
#basecomplement = {'A':'T', 'T':'A', 'G':'C', 'C':'G', 'N':'N'}
basecomplement = { 
       'A' : 'T',
       'T' : 'A',
       'C' : 'G',
       'G' : 'C',
       'n' : 'n',
       'N' : 'N',
       'R' : 'Y',
       'Y' : 'R',
       'M' : 'K',
       'K' : 'M',
       'W' : 'W',
       'S' : 'S',
       'B' : 'V',
       'V' : 'B',
       'D' : 'H',
       'H' : 'D',
}
def reverse_complement(sequence): 
    """
    Reverse complement of a sequence
    
    sequence  --  DNA string
    
    Return    -- reverse complement of the dna string.
    """ 
    letters = list(sequence) 
    #letters.reverse() 
    letters = [basecomplement.get(base, base) for base in letters[::-1]] 
    return ''.join(letters) 


def is_valid(name, location, sequence):
    """
    Test sequence validity
    
    sequence     -- string sequence
    
    location     -- tuple [start, end]
    
    new_location -- tuple [new_start, start]

    Return       -- True if sequence is in [A,T,G,C] and length correspond to location
    """
    #name, location, sequence = input
    if location[1] - location[0] +1!= len(sequence):
        return False
    for base in sequence:
        if base not in ['A', 'C', 'G', 'T', 'N']:
            return False
    return True
#
# GC
#
def gc_percent(dna):
    """
    Compute gc%
    
    dna    -- DNA strings
    
    Return -- float
    """
    return (dna.count('G') + dna.count('C')) / float(len(dna))
    
def ACGT_freq_dna(sequences, start=None, end=None):
    """
    End is excluded
    
    sequences -- list of dna string

    Optional
    start     -- start in [0..len(seq)]
    end       -- start in [0..len(seq)]
    """
    A,C,G,T = 0,0,0,0
    l = 0
    for s in sequences:
        A += s.count('A', start or 0, end or len(s))
        C += s.count('C', start or 0, end or len(s))
        G += s.count('G', start or 0, end or len(s))
        T += s.count('T', start or 0, end or len(s))
        l += len(s)
    if start != None and end != None:
        L = float(end - start) * len(sequences)
    else:
        L = float(l)
    return [A/L,C/L,G/L,T/L]


def ACGT_freq(sequences, start=None, end=None):
    """
    End is excluded
    
    sequences -- list of sequences

    Optional
    start     -- start in [0..len(seq)]
    end       -- start in [0..len(seq)]
    """
    A,C,G,T = 0,0,0,0
    l = 0
    for s in sequences:
        A += s.sequence.count('A', start or 0, end or len(s.sequence))
        C += s.sequence.count('C', start or 0, end or len(s.sequence))
        G += s.sequence.count('G', start or 0, end or len(s.sequence))
        T += s.sequence.count('T', start or 0, end or len(s.sequence))
        l += len(s.sequence)
    if start != None and end != None:
        L = float(end - start) * len(sequences)
    else:
        L = float(l)
    return [A/L,C/L,G/L,T/L]

#
# Misc
#

locationStrM  = re.compile(r'([\+\-]{0,1}\d+):([\+\-]{0,1}\d+)')
locationStrM2 = re.compile(r'from ([\+\-]{0,1}\d+) to ([\+\-]{0,1}\d+)')

def decode_ref(str, start=0, end=0):
    """
    decode UCSC reference
    
    str -- sequence title

    RETURN: tuple (start, end, title)
    """
    try:
        title = str.split(',')[0].split()[0]
    except:
        title = str


    r = locationStrM.search(str)
    nm = None
    if r:
        (start, end) =  int(r.group(1)), int(r.group(2))
        # NM_ ?
        r = re.search(r'(NM_\d{6})', str)
        if r:
            nm = r.group(1)
        return (title, nm, start, end)

    r = locationStrM2.search(str)
    if r:
        (start, end) =  int(r.group(1)), int(r.group(2))            
        return (start, nm, end, title)

    return (title, nm, start, end)


def IUPAC2list(w):
    """
    convert a word in IUPAC code into a list
    """
    l = ['']
    for code in w:
        nl = []
        for x in l:
            if code == 'N':
                nl.append(x + 'A')
                nl.append(x + 'C')
                nl.append(x + 'G')
                nl.append(x + 'T')
            else:
                nl.append(x + code)
        l = nl
    return l


def _extendn(w, NExtension, l):
    i = w.find('n')
    if i == -1:
        l.append(w)
    else:
        for e in range(NExtension[0], NExtension[1]+1):
            _extendn(w[:i] + 'N' * e + w[i+1:], NExtension, l)


def extendn(w, NExtension):
    l = []
    _extendn(w, NExtension, l)
    return l


def _convert_N(word):
    if word[0] == 'N':
        l = ['A', 'C', 'G', 'T']
    else:
        l = [word[0]]

    for i  in range(1,len(word)):
        l2 = []
        for prefix in l:
            if word[i] == 'N':
                l2 += [ prefix + 'A']
                l2 += [ prefix + 'C']
                l2 += [ prefix + 'G']
                l2 += [ prefix + 'T']

            else:
                l2 += [ prefix + word[i] ]

        l = l2
    return l

def convert_N(words):
    assert(type(words) == list)
    r = []
    for word in words:
        r += _convert_N(word)
    return r



def IUPAC2fulllist(w, NExtension):
    l = []
    _extendn(w, NExtension, l)
    r = []
    for w in l:
        r += IUPAC2list(w)
    return r


##########################################################
# Method to read a multifasta file and create a serialized
# dump of the sequences. 
# Used to store the databases of promoter sequences
# Used in the calculation of the background model and in tfme
# itself to find the sequence belonging to an identifier.
##########################################################
def fasta2dump(fastaFile,organism,location,dumpfilename):
    """
    create a dump of Sequences taken from a fasta file. Read the fasta file 
    extracts the sequences and create a dump file.

    fastaFile    -- a multifasta file containing the sequences
    
    organism     -- the organism of all the sequences in the file
    
    location     -- the location of all the sequences in the file
    
    dumpfilename -- the file where to store the binary mode of the Sequences list
    """
    try:
        p = op.P(dumpfilename, 'c') #create the shelf and store it in a binary file
       
        
        p['%location'] = location
        p['_info'] = 'SEQUENCES'
        l = []

        ###################################
        # check the format of the fasta file        
        inputfile  = Core.IO.utils.uopen(fastaFile, mode='r')
        #seq=ic.format_fasta(inputfile)
        
        if ic.verif_fasta(inputfile) == False:
            cli.error("The fasta file has incorrect format. Please check your inputs")
            sys.exit(1)

        info = cli.Info(length=None,showTime=False, showRemaining=False,verbosity=1)
        
                
         
        ################################                
        ##find the smallest dna sequence
        uf=file(fastaFile,'r')
        strike=0
        smallestDna=0
        for title1,dna1 in fasta.FastaReader(uf):
            
            if strike == 0:
                smallestDna=len(dna1)
            else:
                if len(dna1) < smallestDna:
                    smallestDna=len(dna1)
            
            strike+=1
        
        #####################
        ##choose the location
        if location:
            lenLoc=(location[1]-location[0])+1
            if lenLoc<=smallestDna:
                loc=location
            else:
                loc=[-smallestDna+1,0]
                cli.warning("Some sequences were shorter than the entered location. The location has been changed to "+str(loc)+"\n")
        else:
            loc=[-smallestDna+1,0]
            cli.warning("The location has been changed to "+str(loc)+"\n")  
          

        ##################################
        # extracts sequences    
        uf2=file(fastaFile,'r')
        for title, dna in fasta.FastaReader(uf2):
                dna = dna.upper()
                id,org,loca = decode_title(title)
                if org != organism.lower():
                    cli.warning('An organism in the fasta file is not the same as the one specified in the command: '+title)
                
                lenLocation=(loc[1]-loc[0])+1
                removePart=len(dna)-lenLocation
                if removePart >0:
                    dna=dna[removePart:]

                if is_valid(id, loc, dna):
                
                    seq=Sequence(id, loc, dna, organism=organism or org)                    
                    p[id]=seq
                    l.append(id)

                else:
                    cli.warning('Invalid sequence %s' % (title))
                            
                info('Building dump for %s ' % (id))
         
        p.keys()
               
        organisms = p.get('%organisms', [])
        organisms.append(organism)
        p['%organisms'] = organisms

        ll = p.get('%'+organism, [])
        ll.extend(l)
        p['%'+organism] = ll
    
        #op.osave(p,dumpfilename)

    except IOError:
	    cli.error("An error occured in creating the dump file")
    
##########################################################
        
#-------------------------------------------------------------------------------
# CODE



