####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Bio.transfac
# DESC Utilities for reading TRANSFAC files
#                                                                              #
################################################################################
"""
:NAME:
Bio.transfac

:DESC:
Utilities for reading TRANSFAC files

classes:

methods:
    read_sections
    get
    read_gene
    read_matrix

"""
# IMPORTS
#-------------------------------------------------------------------------------
import re
from StringIO import StringIO

from Core.Types.string import find
from Core.IO.section import SectionIterator
#-------------------------------------------------------------------------------
# IMPORTS


# CODE
#-------------------------------------------------------------------------------
def read_section(s, keep=None):
    """
    return dict of string
    """
    keep = keep or []

    l = s.splitlines()

    r = dict([(k, '') for k in keep])

    for i in l:
        if i[:2] in keep:
            r[i[:2]] += i[2:].strip()
    return r


def get(line):
    """
    Extract the information from a line of transfac file
    
    line -- a line of file
    
    return -- a string 
    """
    return line[:2], line[2:].strip()


def read_gene(s):
    """
    Read TRANSFAC gene from string
    
    s -- string
    """
    keep = ['OS']
    

    r = {
    'OS'    : '',
    'TF'    : [],
    'BS'    : [],
    'UniGene': '',
    'RefSeq' : '',
    
    
    }

    d = read_section(s, ['OS', 'BS', 'DR'])
    #OS
    r['OS'] = d['OS'].split(',')[0].strip()

    #DR
    for x in d['DR'].split('.')[:-1]:
        if x.split()[0] == 'UNIGENE:':
            r['UniGene'] = find(r'\w\w\.\d+', x)
        elif x.split()[0] == 'REFSEQ:':
            r['RefSeq'] = find(r'NM_\d{6}', x)        
    #BS
    for x in d['BS'].split('.')[:-1]:
        TFList = re.findall(r'<(T\d+)>', x)
        if TFList:
            r['TF'] += TFList
            pos = re.findall(r'\s*([+-]?\d+)\s+([+-]?\d+).+', x)
            if pos:
                for TF in TFList:
                    r['BS'].append( (pos[0][0], pos[0][1], TF) )
    
    return r


pattern_0X = re.compile(r'^\d{2}')

def read_matrix(s):
    """
    Read the content of a matrix public Transfac file and extract the matrix itself
    
    s      -- a string with a part of Transfac file
    
    return -- a dictionnary with the matrix
    """
    fieldsToKeep = ['AC', 'ID', 'NA', 'DE']

    d= {}
    d['matrix'] = matrix = []
    d['BF'] = []
    d['BA'] = 28
    d['species'] = []
    
    try:
        for i in s.splitlines():
            if i[:2] in fieldsToKeep:
                d[i[:2]] = i[2:].strip()
            elif i[:2] == 'BA':
                d['BA'] = int(find(r'.*?\s(\d+\.\d+|\d+)\s.*', i) or '28')
            elif i[:2] == 'BF':
                d['BF'] += [ i[2:].split(';')[0].split()[0].strip() ]
                try:
                    spe = i[2:].split(';')[1].split(':')[1].split(',')[0].strip()
                except:
                    spe = ''
                d['species'].append(spe)
            elif i[:2] == 'P0' or i[:2] == 'PO':
                d['bases'] = i[2:].split()
            elif pattern_0X.search(i[:2]):
                matrix += [ i[2:].split()[:4] ]
                                
            #elif find(r'^\d{2}', i[:2]):
            #    matrix += [ i[2:].split()[:4] ]    

        return d
    except:
        cli.error('The Transfac file is malformed: '+s)
        sys.exit(1)


def is_included(dict, indict):
    for key, value in dict.items():
        if not indict.has_key(key) or indict[key] != value:
            return False
    return True

gene_dat = '/export/data/TRANSFAC/8.1/Dat/gene.dat'

def co_regulated_genes(f, OS, TFs):
    #OS = 'human'
    #TF  = 'T00029'
    f = SectionIterator(f)
    f.next()

    l = []

    res = {}
    for s in f:
        d = read_gene(s)
        for TF in TFs:
            if d['OS'] == OS and (TF in d['TF']):
                #print d['UniGene'], '\t', d['RefSeq']
                if d['RefSeq']:
                    print d['RefSeq']
                    l += [ d['RefSeq'] ]
                    #print d['RefSeq'], [i for i in d['BS'] ]
                    #print d['BS']
                    #print d['TF']
                    res[d['RefSeq']] =  [i for i in d['BS'] if i[-1] == TF]
    return res


def true_sites(f, OS, nms, location):
    #OS = 'human'
    #TF  = 'T00029'
    f = SectionIterator(f)
    f.next()

    l = []

    res = {}
    for s in f:
        d = read_gene(s)
        if d['OS'] == OS and d['RefSeq'] in nms:
            for bs in d['BS']:
                TF = bs[-1]
                if int(bs[0]) > location[0] and int(bs[0]) < location[1]:
                    res[TF] =  res.get(TF, []) + [ (bs[0], d['RefSeq']) ]
    return res


#####  remove this
import Core.types as types
def count(res, tfs):
    c = 0
    for tf in tfs:
        if res.has_key(tf):
            c += len(res[tf])
    return c


def get_sorted(res, tfs):
    error = 100

    
    h = {}
    for tf in tfs:
        if res.has_key(tf):
            for i in res[tf]:
                h[i[1]] = h.get(i[1], []) + [ int(i[0]) ]

    for k in h:
        h[k] = types.list.sorted( [(i-error, i+error) for i in  h[k]] ) 

    return h



###############


def count_tfbs_in(res, location, tfs):
    count = 0

    positionError = 5
    

    for v in res.values():
        for i in v:
            position = int(i[0])
            if position >= location[0]-positionError and position <= location[1] + positionError and i[2] in tfs:
                count += 1
    return count



def get_all_nm(f, OS, matrices):
    l = []
    for m in matrices:
	print m.id
	for bf in m.bf:
	    print bf
	    l += co_regulated_genes(open(f), OS, bf)
    return l

factor_dat = '/export/data/TRANSFAC/8.1/Dat/factor.dat'
def read_factors(f):
    """
    return dict
    key = factor name
    value = list of tuple (AC, OS) #accession, organism
    """

    f = SectionIterator(f)
    f.next()    

    all = {}
    for s in f:
        d = read_factor(s)
        all[d['FA']] = all.get(d['FA'], [])
        all[d['FA']] += [ (d['AC'], d['OS']) ]
    return all


def read_factor(s):
    fieldsToKeep = ['AC', 'FA']

    d= {}
    for i in s.splitlines():
        if i[:2] in fieldsToKeep:
            d[i[:2]] = i[2:].strip()
        if i[:2] == 'OS':
            d['OS'] = i[2:].split(',')[0].strip()

    return d



#-------------------------------------------------------------------------------
# CODE
