####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Types.list
# DESC List utilities
#                                                                              #
################################################################################
"""
:NAME:
Types.list

:DESC:
List utilities and SotedList implementation

"""

# IMPORTS
#-------------------------------------------------------------------------------
from copy import copy
from operator import add
#import sets
#-------------------------------------------------------------------------------
# IMPORTS

# OBJECTS
#-------------------------------------------------------------------------------

class SortedList(list):
    """A simple sorted list
    """

    def __init__(self, max=10, growing=True):
        """
        max       -- max length
        growing   -- growing order? Boolean (True = [1,2,3] or False = [3,2,1]
        """
        list.__init__(self)
        self.max     = max
        self.growing = growing

    def inf(self, a, b):
        return a <= b

    def sup(self, a, b):
        return a >= b    

    def append(self, elt):
        # try to append element
        i = 0
        try:
            while 1:
                if self.growing:
                    if self.inf(elt,self[i]):
                        break
                else:
                    if self.sup(elt,self[i]):
                        break
                i += 1
            self.insert(i, elt)
            if len(self) > self.max:
                self.pop()

        except IndexError:
            if len(self) < self.max:
                list.append(self, elt)


    def extend(self, other):
        for i in other:
            self.append(i)

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __add__(self, l):
        raise NotImplementedError

#-------------------------------------------------------------------------------
# OBJECTS

# CODE
#-------------------------------------------------------------------------------

def sum(l):
    return reduce(add, l)

def avg(l):
    return reduce(add, l) / float(len(l))
  
def freq(l):
    s = sum(l)
    return map(lambda x: x/s, l)

#def flat(l):
#    """
#    Example: flat([[1,2], [3,4]]) return [1,2,3,4]
#    """
#    return reduce(lambda x,y:x+y, l)

def flatten(l):
    r = []
    for i in l:
        if type(i) is tuple or type(i) is list:
            r.extend(i)
        else:
            r.append(i)
    return r


def chunked(l, chunkSize=10):
    """Return Chunks of l of length chunkSize
    chunkSize -- int
    Example chunked([1,2,3,4], 2) return [[1,2],[2,3]]
    """
    r = []
    for i in range(0, len(l) - chunkSize+1, chunkSize):
        r.append(l[i:i+chunkSize])
    if (len(l) % chunkSize) != 0:
        r.append(l[-(len(l) % chunkSize):])
    return r


def unzip(l):
    """
    """
    if len(l) == 0: 
        return ()
    r = []
    for i in range(len(l[0])):
        r.append(map( lambda e: e[i], l ) )
    return tuple(r)



def select(l, selectFunction):
    """
    l              -- list
    selectFunction -- function that take an item as argument and return True if item is selected or False

    return list of selected items
    """
    
    return [ i for i in l if selectFunction(i) ]
    #r = []
    #for i in l:
    #    if selectFunction.__call__(i):
    #        r.append(i)
    #return r


def count(l, condition):
    c = 0
    for e in l:
        if condition(e):
            c += 1
    return c


def select_first(l, selectFunction):
    """
    l              -- list
    selectFunction -- function that take an item as argument and return True if item is selected or False

    return first selected item or None
    """
    for i in l:
        if selectFunction(i):
            return i
    return None


def top(l, topSize=-1):
    """Return topSize element of l
    
    topSize -- default -1 all list
    """
    if topSize == -1:
        return l
    else:
        return l[:min(topSize, len(l))]


def unique(l):
    return list(set(l))


def elements_index(l, f):
    """Index list l with function f
    Note: f  must return a unique item id
    Return dict that associate e to key f(e)
    """
    return dict( [ (f(e), e) for e in l ] )


def indicies_index(l, f):
    """Index list l with function f
    Note: f  must return a unique item id
    Return dict that associate position in l to key f(e)
    """
    return dict( [ (f(e), l.index(e)) for e in l ] )


def l2categories(l, f):
    """Categorize a list
    l -- list
    f -- function that take a list item and return is category

    return l where elements are categories
    """
    return [ f(e) for e in l ]


def categorize(l, f):
    """Categorize a list
    l -- list
    f -- function that take a list item and return is category
    
    #lambda i: getattr(i, attr)

    return dict key=category, value=list of item that are in this category
    """
    categories = {}
    for i in l:
        categories[f(i)] = categories.get(f(i), []) + [ s ]
    return categories

def categorize_idx(l, f):
    """Categorize a list
    l -- list
    f -- function that take a list item and return is category
    
    #lambda i: getattr(i, attr)

    return dict key=category, value=list of item index in l that are in this category
    """
    categories = {}
    for i in l:
        categories[f(i)] = categories.get(f(i), []) + [ l.index(i) ]
 
    return categories


def sorted(l, cmpfunc=None):
    """Return sorted copy of the list
    """
    x = copy(l)
    x.sort(cmpfunc)
    return x

def reversed(l):
    """Return reversed list
    """
    x = copy(l)
    x.reverse()
    return x

    
def trimed(l, cond=(lambda v:1)):
    """ Return a new list in which cond(v) == true items are removed.
    l    -- list to trim
    cond -- function with value args
    """
    return [i for i in l if not cond(i)] 


def find(l, condition):
    """
    l -- sorted list
    search in l for first item where condition(item) is true
    return is index or None
    """
    for i in range(0,len(l)):
        if condition(l[i]):
            return i
    return None


def get(l, index, default=None):
    try:
        return l[index]
    except IndexError:
        return default


########################################
#                                      #
#         2 lists opetations           #
#                                      #
########################################


def correlation(l, other):
    """
    compute correlation between l and other

    return a number n 0 < n < 1
    0 if no element in l are in other
    1 if all elements in l are in other

    """
    
    assert(len(l) == len(other))
    X = 0
    for i in range(len(l)):
        if l[i] == other[i]:
            X += 1
    return float(X) / len(l)



def diff(l, l2):
    """return l - l2 (elements of l that are not in l2)
    """
    return [i for i in l if not i in l2]

def intersection(l, l2):
    return list(set(l).intersection(set(l2)))


def union(l, l2):
    return list(set(l).union(set(l2)))

def is_subset(l, l2):
    """
    return True is all elements of l are in l2
    """
    for e in l:
	if e not in l2:
	    return False
    return True



########################################
#                                      #
#             lt2uf, u2lf              #
#                                      #
########################################
def uf2l(f):
    """Convert simple file format(one item per line) to list
    skip lines that start with # 
    f  -- file or filename
    return list
    """
    if type(f) is str:
        f = open(f, 'r')    
    l = []
    for line in f:
        line = line.strip()
        if not line.startswith('#') and len(line) > 0:
            l.append(line)
    return l


def l2uf(l, f):
    """Convert list to simple file format (one item per line)
    l  -- list
    f  -- file or filename
    """
    if type(f) is str:
        f = open(f, 'w')
    f.write('\n'.join(map(str,l)))
    f.close()

#-------------------------------------------------------------------------------
# CODE
