####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Types.pipe  
# DESC Iterator pipe
#                                                                              #
################################################################################
"""
:NAME:
Types.pipe

:DESC:
Iterator pipe

:LONG_DESC:
unix pipe like
Transposition of unix pipe object to python Oject

class to inherit:
(all class are Iterators)
-Input  every pipe chain must begin with a subclass of this
-Output every pipe chain must end with a subclass of this
-Filter every inner pipe chain element must be a subclass of this

Input:   must overload action(self) method
Output:  must overload action(self, input) method
Filter:  must overload action(self, input) method

Simple Example of subclass:
-FileReader
-FileWriter
-UpperCase

Implemented filters:
Counter([count])
SelectRange(i,j) : keep items in [i:j]
Apply(filter)    : apply filter (filter is a function that take an Oject and return an Object)

Implemented Output:
Print()
NULL()

:EXAMPLE:
FileReader(open('txt')) | UpperCase() | FileWriter(open('out', 'w'))


"""
# IMPORTS
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# IMPORTS

# OBJECTS
#-------------------------------------------------------------------------------
class Input:
    """
    Generic Input class
    Method to overload: action    
    """
    def __init__(self, action=None):
        """
        action  -- function
        """
        if action:
            self.action = action

    def __iter__(self):
        return self

    def next(self):
        r = self.action()
        if r == False:
            raise StopIteration
        else:
            return r

    def action(self):
        """
        overload this method
        return object or False
        """    
        pass


class Output:
    """
    Generic Output class
    Method to overload: action
    NOTE:
    if the action return an object this object is stored in a list whitch is returned after execution
    """
    def __init__(self):
        pass

    def __ror__(self, pipein):
        l = []
        for i in pipein:
            output = self.action(i)
            if output != None:
                l += [output]
        if l:
            return l

    def action(self, input):
        """
        overload this method
        can return an object
        """
        return input

class Filter:
    """
    Generic Filter class
    Method to overload: action
    """
    def __init__(self):
        pass

    def __iter__(self):
        return self

    def skip(self):
        self.pipein.next() 

    def next(self):
        return self.action(self.pipein.next())

    def __ror__(self, pipein):
        self.pipein = pipein
        return self
        
    def action(self, input):
        """
        overload this method
        """
        return input.upper()

##
#
# Example
#
##
class FileReader(Input):
    """
    A simple Input example
    """
    def __init__(self, file=None):
        Input.__init__(self)
        self.f = file

    def action(self):
        l =  self.f.readline()
        if not l:
            return False
        return l

    def __getitem__(self, filename):
        self.f = open(filename)
        return self

read = FileReader()


class FileWriter(Output):
    """
    A simple Output example
    """
    def __init__(self, file):
        Output.__init__(self)
        self.f = file

    def action(self, input):
        self.f.write(input)
        self.f.flush()


class UpperCase(Filter):
    """
    A simple Filter example
    """
    def __init__(self):
        Filter.__init__(self)

    def action(self, input):
        return input.upper()

##
#
# Utils for pipe
#
##
class Print(Output):
    """
    Output that print received object
    """
    def __init__(self):
        """
        """
        Output.__init__(self)

    def action(self, input):
        print input,




class ApplyToOutput(Output):
    """
    Output Filter
    """
    def __init__(self, action):
        """
        action -- function(input) -> return output
        """
        
        Output.__init__(self)
        self.act = action

    def action(self, input):
        """
        """
        return self.act(input)


class Apply(Filter):
    """
    An apply filter facilitie
    """
    def __init__(self, filter, *args):
        """
        filter -- function(input, *args) -> return output
        *args  -- args
        """
        Filter.__init__(self)
        self.filter = filter
        self.args = args
        
    def action(self, input):
        self.r = self.filter(input, *self.args)
        return self.r

class NULL(Output):
    """
    A NULL Output
    """
    def __init__(self):
        Output.__init__(self)
    def action(self, input):
        pass

class SelectRange(Filter):
    """
    The select range filter
    """
    def __init__(self, begin=0, end=10):
        """
        """
        Filter.__init__(self)
        self.begin   = begin
        self.end     = end
        self.current = -1

    def action(self, input):
        self.current += 1
        if self.current > self.end:
            raise StopIteration
        if self.current < self.begin:
            while self.current < self.begin:
                input = self.pipein.next()
                self.current += 1
        return input
        
    def __getslice__(self, i, j):
        self.begin = i
        self.end   = j
        return self


class Select(Filter):
    def __init__(self, test):
        """
        test -- test function return True/False
        """
        Filter.__init__(self)
        self.test   = test

    def action(self, input):
        if not self.test(input):
            input = self.pipein.next()
            return input
        else:
            raise StopIteration

class Counter(Filter):
    """
    Simple Counter
    """
    def __init__(self, count=None):
        """
        Simple counter:
        usage: instantiate, use, call oject -> counter_value or instantiate with count a list [0]
        and get count value with this list
        
        count -- optional list of one int where to store the count [0]
        """
        self.count = count or [0]
        Filter.__init__(self)
        self.value = 0

    def action(self, input):
        self.value += 1
        self.count[0] += 1
        return input

    def __call__(self):
        return self.value
        
class CounterRunner(Filter):
    def __init__(self):
        """
        """
        Filter.__init__(self)
        self.current = 0

    def action(self, input):
        self.current += 1
        #utils.runner("[%d]" % self.current)
        return input

#-------------------------------------------------------------------------------
# OBJECTS
  

