####################
# version 2.0
####################


"""
 Toolbox with facilities to check user inputs in tfme

 Authors: Laurie Tonon

 Date: 28/11/2009
 
 methods:
    format_refseq
    verif_fasta
    format_fasta
    verif_fasta_content
    verif_jaspar
"""



import re    
import Bio   
                
def format_refseq(seq):
	"""
	read a list of NM identifiers and format them by removing ignored characters
	
	seq    -- string with list of NM id
	
	return -- a string with the corrected NM identifiers
	"""
	lines = seq.readlines()
	res=[]

	for line in lines:
		line=line.strip()
		if line.startswith('>') or line.startswith('#'):
			continue
		else:
		
			p=re.compile(r'^NM_(\d)+',re.I)
			if re.search(p,line):
				res.append(line)
				
	return '\n'.join(res)

def verif_fasta(seq):
    """
    read a multifasta and check its format
    
    seq    -- a string with a list of fasta sequences
    
    return -- True if the fasta is correct, False otherwise
    """

    seq = seq.read()
    

    if seq is None:
        return False

    pattern =  """
        	^                               # beginning of string
        	(                               # one sequence 
       	 	\ *>.+[\r\n]+                   # line of the name
        	([ACGTNacgtn]+[\r\n]*)+      # line of the dna sequence
        	)                               # end of one sequence
        	{1,}$                           # at least one sequences
        	"""

    if re.search(pattern, seq, re.VERBOSE) is None:
        return False
    else:
        return True



def format_fasta(seq):
    """
    read a mulitfasta and format it by removing ignored characters
    
    seq    -- a string with a list of fasta sequences
    
    return -- a string with correted fasta or None
    """
    
    try:
        list_seq = seq.readlines()
        temp_seq = ''

        # remove spaces for non-title line
        for x in list_seq:

            if re.search(r'^ *>', x) is None:
                x = x.lower()
                x = x.strip()
                #dna=[base for base in x if base in ['a','c','g','t','u','n']]
                #curated=''.join(dna)
                temp_seq = temp_seq + x + '\n'
            else:
                temp_seq = temp_seq + x + '\n'
         

        return temp_seq
    except:
        return None
        
def verif_fasta_content(seq, seqDB=None):
    """
    check a fasta file to be sure the header is in format: >sedID|org|other 
    
    seq    -- the content of a multi fasta
    
    seqDB  -- a database of sequences (used to retrieve an organism from a sequence id)
    
    return -- an error message and the involved sequences id
    """
    
    
 
    error=False
    err_seq="The following sequences have incorrect location (it does not correspond to the entered location): \n"
    for title, dna in Bio.fasta.FastaReader(seq):
        dna = dna.upper()
        
        
        pattern=".{0,}|.{0,}|.{0,}"
        pattern2=".{0,}|.{0,}"
        pattern3=".{0,}"

        if re.search(pattern,title) is None and re.search(pattern2,title) is None and re.search(pattern3,title) is None:
            error=True
            err_seq="A fasta header is not correct. It must be in the format: '>sequence_id,organism'" 
            break
            
        id, org, other = Bio.sequence.decode_title(title)

        if id is None or id == "":
            error=True
            err_seq="a fasta header is not correct, cannot find sequence id."
            break
        if org is None or org == "":
            if seqDB:
                if Bio.sequence.find_organism(id,seqDB) != None:
                    org= Bio.sequence.find_organism(id,seqDB)
            else:
                error=True
                err_seq="a fasta header is not correct, the organism is unknown or unspecified."  
                break
                  
    return (error,err_seq)
    
def verif_jaspar(mat):
    """
    check that a string is in Jaspar format (one line for one base)
    
    mat    -- a line of a Jaspar matrix
    
    return -- True if pattern match, False otherwise
    """

    if mat is None:
        return False
    
    pattern =  """
        	^                               # beginning of string
       	 	\s*\d+(\s*\d*)*            # line of frequency
            $
        	"""

    if re.search(pattern, mat, re.VERBOSE) is None:
        return False
    else:
        return True
        
