####################
# version 2.0
####################


"""
Window objects and Windows list

1. Extract windows with hit over representation
2. Evaluate an sort those windows

classes:
    Window
    WindowList
   
methods:
    add_info2hits

"""
from copy import copy
import Core.Types.list
import Core.Types.dict
import pvalue

import Bio.sequence
########################################
#                                      #
#        Window / Window list          #
#                                      #
########################################




class Window:
    """
    Window with hit over representation
    """
    def __init__(self, i, j, matrix, pvalue=0.0, evalue=0.0):
        self.i = i
        self.j = j
        self.location = (i,j)
        self.matrix   = matrix
        self.pvalue   = pvalue
        #self.evalue   = evalue

    def __cmp__(self, other):
        return cmp(self.pvalue, other.pvalue)

    def str(self):
        d = copy(self.__dict__)
        d['matrix'] = `self.matrix`
        return '%(matrix)-20s [%(i)+.4d:%(j)+.4d] %(pvalue).3e' % d

    def to_dict(self):
        """
        Convert a Window object into a dictionnary.
 
        """

        return {'i':self.i,
                'j':self.j,
                'location' : (self.i, self.j),
                'locationStr' : "[%+05d:%+05d]" % (self.i, self.j),
                'matrixId' : self.matrix.id,
                'matrixName' : self.matrix.family,
                'matrixInfo': self.matrix.info,
                'matrixGC': self.matrix.gc,
                'matrixLen': len(self.matrix),
                'pvalue'     : self.pvalue,
                #'evalue'     : self.evalue,
                'numberOfHits' : self.numberOfHits,
                'numberOfSequences' : len(self.sequencesIn),
                'percentOfSequences': self.percentOfSequences * 100.0,
                'sequences' : [s.id for s in self.sequences],
                'sequencesName' : self.sequences.name,
                'sequencesIn' : [s.id for s in self.sequencesIn],
                'sequencesVector' : self.sequencesVector,
                'sequencesGC' : self.sequencesGC,
                'hits'      : self.hits,
                'hitsList'   : self.hitsList,
                'hitsFull'   : self.hitsFull,
                'hitsStats' : self.hitsStats,
                'score'      : self.score,
             }
    
    def full_str(self):
        return windowdict2str(self.to_dict())

    def latex_str(self):
        d = copy(self.to_dict())
        d['matrixId'] = d['matrixId'][2:].replace('_', '\\_')
        s = '\\tt %(matrixId)-20s & %(matrixName)-15s & \\tt [%(i)+05d:%(j)+05d] &' % d
        s += '%(numberOfHits)4d & %(percentOfSequences)3d & %(pvalue).3e' % d
        return s

    def __repr__(self):
        return self.full_str()


def add_info2hits(l, sequences, matrix):
    """ 
    Create a string representing a hit for a matrix and sequences
    
    l         -- list of hitsStr
    
    sequences -- user sequences
    
    matrix    -- a matrix used to scan the sequences
    
    return    -- an array of string, each string being a hit
    """
    r = []
    r += [ '#seq\tabs_pos\trel_pos\tstrand\tpval\tsite' ]
    for i in range(len(l)):
        if len(l[i]) == 0:
            continue
        #r += [str(i[0][0]) + '\n\t' + ' '.join(['%+05d' % h[1] for h in i])]
        for h in l[i]:
            s = sequences[i]
            word = s.sequence[h[1]:h[1]+len(matrix)]
            if h[2] == '-':
                word = Bio.sequence.reverse_complement(word)
            r += [ '%s\t%05d\t%+05d\t%s\t%5.2f\t%s' % (h[0], h[1], h[2], h[3], h[4], word)]

    return r


class WindowList(Core.Types.list.SortedList):
    """Window list sorted by P-value
        
    """
    def __init__(self, title='', maxSize=20, params=None):
        super(WindowList, self).__init__(maxSize)
        self.title  = title
        self.info   = ''
        self.params = params 

    def set_info(self, h):
        """
        add information about hits to the window (nb of hits, hits per sequences ...)

        h -- hits list
        """
        for i in range(len(self)):
            w = self[i]
            w.sequencesVector = h.sequences_vector(w.matrix, w.location)
            w.sequences = h.sequences
            w.sequencesIn = [w.sequences[j] for j in range(len(w.sequences)) if w.sequencesVector[j]]
            w.numberOfHits = h.count_hits_in(w.matrix, w.location)
            w.percentOfSequences = len(w.sequencesIn) / float(len(w.sequences))
            w.sequencesGC = Core.Types.list.avg([ s.gc() for s in w.sequences])

            w.hits = h.select(w.matrix, w.location)
            w.hitsList = h.selection2formatedlist(w.hits)
            w.hitsFull = add_info2hits(w.hitsList, w.sequences, w.matrix)
            #hits distribution
            
            hitsPerSeq = [len(x) for x in w.hits]

            w.hitsStats = { 'hitsPerSeq' : hitsPerSeq,
                            'avg1bp' : sum(hitsPerSeq) / float(len(w.sequences)) / float(w.location[1] - w.location[0] +1),
                            'avg1bpin' : sum(hitsPerSeq) / float(len(w.sequencesIn)) / float(w.location[1] - w.location[0] +1)
                            }
            
    def full_str(self):
        return list2str([i.to_dict() for i in self])
                    

    def __repr__(self):
        s = []
        for i in range(len(self)):
            s += ['[%2d] ' % (i+1) + `self[i]`]
        return '\n'.join(s)


    def latex_str(self):
        s = []
        for i in range(len(self)):
            s += [ self[i].latex_str() + '  \\\\ ']
        return '\n'.join(s)
        

