#! /usr/bin/env python

###################
# version 2.0
###################

#############################################################################################
# tfme
# Program that runs the TFM-Explorer algorithm. 
#
# Inputs: A fasta file or a file with NM identifiers
#         A background model (dump file from tfmebg)
#         A matrices file (dump file from matrix2dump)
#         If NM identifiers, a sequences file to retrieve the sequences (dump file from fasta2dump)
#         An output directory
#         A location to scan
#         Parameters
#
# Output: On screen: table with cluster identified by TFM-Explorer
#         Files: Clusters, correlations, hits 
#
# Authors:Matthieu DeFrance, modified by Laurie Tonon
# Date: 28/10/2009
#
#############################################################################################

import os, sys
PYTHONPATH = [['TFME'], ['TFME', 'MLib.zlib']]
if sys.version_info < (2, 3):
    print >> sys.stderr, 'python >= 2.3 is required!'; sys.exit()
[sys.path.insert(1, os.path.join(sys.path[0], *path)) for path in PYTHONPATH]


import imp

#********************************************************************
# Check dependencies
#********************************************************************
try:
    import numpy
    version= numpy.__version__

except ImportError:
    print 'WARNING: TFM-Explorer requires Numpy '
    sys.exit(1)
    
try:
    import Pyrex

except ImportError:
    print 'WARNING: TFM-Explorer requires Pyrex '
    sys.exit(1)

import optparse
import glob
from StringIO import StringIO

import Core.Persistence.op as op
import Core.Types.list
import Core.UI.cli as cli
import Core.IO.utils
import Bio.sequence
import sample
import pataud
import time
import outputXML
import input_checker as ic

usage = """
      python  %prog  --input=filename  [--fasta] --background=file.bg --matricesDB=file.mat.db --output= ../resultsTFM/ --location=x:x [--sequencesDB=file.seq.db]  [--top=x or --pvalmax=x] [--ratio=x]
"""

parser = optparse.OptionParser(usage)

parser.add_option("-l", "--location", dest="location", help="region to scan. boundaries must be lower or equal to the ones in the background model  (example -2000:-1) (default= -2000:200)", action="store", type="string", default='-2000:+200')
parser.add_option("-r", "--ratio", dest="ratio", help="ratio used to count against background model. Level from which a motive is considered over-represented (between 1.0-3.0) (default=2.5)", action="store", type="float", default='2.5')
parser.add_option("-i", "--input", dest="input", help="input filename (by default: a list of NM_* id)", action="store")
parser.add_option("-f", "--fasta", dest="fasta", help="To specify if fasta input", action="store_true", default=False)
parser.add_option("-v", "--verbosity", action="store", dest="verbosity", type="int", default=1, help="display complete scan process on screen (between 1 and 3) (default=1)")
parser.add_option("-t", "--top", dest="top", help="show only top elements (max 50) (default=20)", action="store", type="int")
parser.add_option("-p", "--pvalmax", dest="pvalmax", help="show only elements with pvalue < pvalmax (max 50)", action="store", type="float")
parser.add_option("--alignments", dest="alignments", default="", action="store", type="string")
parser.add_option("--debug", dest="debug", action="store_true", default=False , help="show program errors. For developpers only")
parser.add_option("-b", "--background", action="append", dest="background", type="string", default=[], help="list of background models to use (must be bg files produced by tfmebg.py)")
parser.add_option("-o", "--output", action="store", dest="output", help="the output directory where the result files will be put", default="ResultsTFME_"+time.strftime("%d_%m_%y_%H_%M",time.localtime()))
parser.add_option("-s", "--sequencesDB", action="append", dest="sequencesDB", type="string", default=[], help="list of sequences DB filenames if the input is NM identifiers (must be dump files produced by fasta2dump.py")
parser.add_option("-m", "--matricesDB", action="append", dest="matricesDB", type="string", default=[], help="list of matrices DB files to use (must be dump files produced by matrix2dump.py)")

(options, args) = parser.parse_args()
cli.VERBOSITY = options.verbosity

def run(args, options):

    # buffers for errors
    error_found = False
    error_messages = []
    
#******************************************************************
# Process user inputs
#******************************************************************
    
    #########################
    # read and check location
    #########################    
    location = Bio.sequence.str2location(options.location)
    if location[0] == location[1]:
        cli.error("The location is null")
        sys.exit(1)
   
    ##########################################################
    # select pvalmax or top to know how many windows to output
    ##########################################################
    top=20
    pvalmax=""
    if options.top and not options.pvalmax:
        if options.top <=50 :
            top=options.top
    elif options.pvalmax and not options.top:
        top=50
        pvalmax=options.pvalmax
    elif options.top and options.pvalmax:
        if options.top <= 50:	
            top=options.top 
            
    #################################
    # check ratio
    #################################
    if options.ratio < 1.0 or options.ratio > 3.0:
        cli.warning('ratio outside authorized values (1.0-3.0). Put to 2.5')
        options.ratio=2.5

    ################################################
    # read, check and prepare the sequences
    ################################################
    if not options.input:
        error_found=True
        error_messages.append('no input file provided. Cannot continue')
    elif not options.fasta and not options.sequencesDB:
        error_found=True
        error_messages.append('Need a sequences dump file with NM identifiers')
    else:
        sequences=None
        warning=""
        if not os.path.isfile(options.input):
            error_found=True
            error_messages.append("Cannot find input file " +options.input)
        else:
            inputfile  = Core.IO.utils.uopen(options.input, mode='r')
            
            if options.fasta: 
                seq=ic.format_fasta(inputfile)
                seq_test=ic.verif_fasta(StringIO(seq))
                
                if options.sequencesDB:
                
                    sequencesDB=[]
                    for filename in Core.IO.utils.super_glob(options.sequencesDB):
                        if os.path.isfile(filename):
                            dump=op.P(filename)
                            if dump.get('_info') != 'SEQUENCES':
                                cli.warning('file '+ filename +' does not contain sequences. Ignored.')
                            else:
                                sequencesDB.append(dump)
                        elif os.path.isfile(filename+'.db'):
                            dump=op.P(filename+'.db')
                            if dump.get('_info') != 'SEQUENCES':
                                cli.warning('file '+ filename+'.db'+' does not contain sequences. Ignored.')
                            else:
                                sequences_small.append(dump)
                        else:
                            cli.warning("file "+filename+" cannot be found. Ignored.")
                    if seq_test == True:
                        (error,message)=ic.verif_fasta_content(StringIO(seq),sequencesDB)
                        
                        if error == True:
                            error_found=True
                            error_messages.append(message)
                    else:
                        error_found=True
                        error_messages.append("Sequences are not in Fasta format.")
                                    
                    (sequences,warning) = Bio.sequence.fasta2sequences(options.input,location,sequencesDB)
                else:
                    if seq_test:
                        (error,message)=ic.verif_fasta_content(StringIO(seq))
                        if error == True:
                            error_found=True
                            error_messages.append(message)
                    else:
                        error_found=True
                        error_messages.append("Sequences are not in Fasta format.")
                    (sequences,warning) = Bio.sequence.fasta2sequences(options.input,location)

            else: #RefSeq file
            
                seq_tested=ic.format_refseq(inputfile)
                
                if seq_tested is None  or (len(seq_tested) is 0):
                    error_found=True
                    error_messages.append("Sequences are not NM identifiers.")
                else:
                    inputfile=StringIO(seq_tested)
                    
                    if options.sequencesDB:
               	        sequencesDB=[]
               	        for filename in Core.IO.utils.super_glob(options.sequencesDB):
               	            if os.path.isfile(filename):
               	                dump=op.P(filename)
                                if dump.get('_info') != 'SEQUENCES':
                                    cli.warning('file '+ filename +' does not contain sequences. Ignored.')
                                else:
                                    sequencesDB.append(dump)
                            elif os.path.isfile(filename+'.db'):
                                dump=op.P(filename+'.db')
                                if dump.get('_info') != 'SEQUENCES':
                                    cli.warning('file '+ filename+'.db'+' does not contain sequences. Ignored.')
                                else:
                                    sequencesDB.append(dump)
                            else:
               	                cli.warning("file "+filename+" cannot be found. Ignored.")
                                
               	        (sequences,warning) = Bio.sequence.sl2sequences(inputfile, location, sequencesDB)
                    else:
                        cli.error("Cannot find the sequences dump file")
                        raise IOError

            if len(sequences) == 0:
                error_found=True
                error_messages.append("Sequences set is empty.")
                error_messages.append(warning)

        ###################
        #change alignments
        ##################
        if sequences is not None:
            [setattr(s, 'locationTSS', s.location) for s in sequences] # (update obj)
            if options.alignments != '':
                alignments = [int(i) for i in Core.Types.list.uf2l(open(options.alignments))]
                sequences.change_locations(alignments)
           
    
    ######################################
    # read,check and prepare the matrices
    ######################################
    if not options.matricesDB:
        error_found=True
        error_messages.append('no matrices dump file provided. Cannot continue.')
    else:  
        mDB=[]
        matrices=None
        for filename in Core.IO.utils.super_glob(options.matricesDB):
            if os.path.isfile(filename):    
               # mDB.append(op.oload(filename))
               dump=op.P(filename)
                
               if dump.get('_info').find('MATRICES') == -1:
                   cli.warning("file "+ filename+ "does not contain matrices. Ignored")
               else:
                   mDB.append(dump)
            elif os.path.isfile(filename+'.db'):
                dump=op.P(filename+'.db')
                if dump.get('_info').find('MATRICES') == -1:
                    cli.warning('file '+ filename+'.db'+' does not contain matrices. Ignored.')
                else:
                    mDB.append(dump)
            else:
                cli.warning("file "+filename+" cannot be found. Ignored.")
        matrices = Core.Types.list.flatten([d.values() for d in mDB])
        


        if len(matrices) == 0:
            error_found=True
            error_messages.append("Matrices set is empty.")       


    ##############################################
    # read,check and prepare the background models
    ##############################################
    if not options.background:
        cli.error('no background model provided. Cannot continue.')
        sys.exit(1)
        
    bgDB=[]
    for file_bg in Core.IO.utils.super_glob(options.background):
        if os.path.isfile(file_bg):
            bgDB.append(file_bg)
        elif os.path.isfile(file_bg+'.db'):
            bgDB.append(file_bg+'.db')
        else:
            cli.warning("file "+file_bg+" cannot be found. Ignored.")
            
    if sequences is not None and len(sequences) != 0:
        bg = pataud.load_bg(files=bgDB,loc=sequences[0].location)
    else:
        bg = pataud.load_bg(files=bgDB,loc=location)
    
    if str(bg) =="":
        error_found=True
        error_messages.append("Background is empty.")                

    
    #######################
    # load the parameters
    #######################
    DEFAULT_PARAMETERS = {'minSize': 30, 'maxSize': 1500, 'ratio': options.ratio, 'top': top, 'pvalmax': pvalmax}
    parameters = DEFAULT_PARAMETERS

    INFO = ''';Sequencesfilename : %s
;Matrices          : %s
;Background        : %s''' % (str(options.input), options.matricesDB, options.background)



    ##############################
    # create the output directory
    ##############################
    if not options.output:
        cli.warning("The output directory is not specified. The results will be printed in the current directory")
        options.output="ResultsTFME_"+time.strftime("%d_%m_%y_%H_%M",time.localtime())
    try:
        os.makedirs(options.output)
    except OSError:
        if os.path.exists(options.output):
        # We are nearly safe
           	 pass
        else:
            # There was an error on creation, so make sure we know about it
       	    cli.error('Cannot create result files, you seem to need rights to write in the output folder')
   
    
#**************************************************
# output possible errors
#**************************************************
    if error_found:
        print "\n[ERRORS]The following errors have been found: "
        for m in error_messages:
            print m
        sys.exit(1)
        
        
   # if sequences is not None and len(sequences)>0 and matrices is not None and len(matrices)>0:
    #    if len(bg.getWithoutSeq(sequences[0].organism,sequences[0].gc(),matrices[0]).p) < (location[1]-location[0]):
     #       error_found=True
      #      error_messages.append("The input sequences are longer than the ones used to create their background model. Impossible to use this model.") 
       
#***************************************************    
# launch tfm explorer
#***************************************************
    print "TFM-Explorer job running..."
    
    (r,finalHits) = sample.scan(sequences, matrices, bg, options.output, parameters)
    
    
#****************************************************
# output the results on screen
#****************************************************
    print >> sys.stdout, sample.r2txt(r, INFO)
    if options.verbosity >= 2:
        print >> sys.stdout, sample.r2txt_details(r, sequences, matrices)
   
    print warning
    
    if r['R']:
        # calculate correlations 2 by 2
        corr=sample.correlations(r)
   	   
#**************************************       
# output the results in xml files
#**************************************
        outputXML.printxml(r,corr,options.output,finalHits)
        
        print 'TFM-Explorer job successfully ended'

if __name__ == '__main__':
    try:
        run(args, options)
    except:
        cli.error('Error while running TFM-Explorer')
        if options.debug:
            raise




